#ifndef GRAPHICPOOL_HPP_INCLUDED
#define GRAPHICPOOL_HPP_INCLUDED

#include <SDL.h>
#include <SDL_image.h>
#include <SDL_ttf.h>
#ifdef __APPLE__
	#include <OpenGL/gl.h>
#else
	#include <GL/gl.h>
	#ifdef __WIN32__
		#include <GL/glext.h>
	#endif
#endif
#include <vector>
#include <string>

class GraphicPool
{
	public:
	//The destructor. When the program ends, it deletes all of the textures.
		~GraphicPool() { empty(); }

	//Getting an instance of the singleton.
		static GraphicPool& instance() {
			static GraphicPool passInstance;
			return passInstance;
		}

	//A small struct for storing the necessary information on the images stored in the pool.
		struct TexData {
			std::string name;
			GLuint tex;

			TexData (std::string nameIn, GLuint texIn) :name(nameIn), tex(texIn) {}
		};

	//The version of the function for returning the texture ID associated with the image called fileName.
	//If it doesn't already exist, it creates it, by trying to find the file called fileName.
		GLuint loadImage(std::string fileName) {
			for(std::vector<TexData>::iterator check(pool_.begin()); check < pool_.end(); ++check)
				if(check->name == fileName)
					return check->tex;

			pool_.push_back(TexData(fileName, genTexture(IMG_Load(fileName.c_str()))));
			return pool_[pool_.size() - 1].tex;
		}

	//The version of the function for returning the texture ID associated with the image called name.
	//If it doesn't already exist, it creates it, by using the surface passed in.
		GLuint loadImage(std::string name, SDL_Surface* surface) {
			for(std::vector<TexData>::iterator check(pool_.begin()); check < pool_.end(); ++check)
				if(check->name == name)
					return check->tex;

			pool_.push_back(TexData(name, genTexture(surface)));
			return pool_[pool_.size() - 1].tex;
		}

	//Empties the graphics out of the pool.
		void empty() {
			for(std::vector<TexData>::iterator kill(pool_.begin()); kill < pool_.end();) {
				glDeleteTextures(1, &(kill->tex));
				kill = pool_.erase(kill);
			}
		}

	protected:
		GraphicPool () { pool_.clear(); }

		std::vector<TexData> pool_;

	private:
	/*
	The function to convert an SDL_Surface* into a glTexture and return the texture's ID.

	Takes 1 parameter:
	surface -> an SDL_Surface pointer that contains the image to be converted to a glTexture.

	Returns 1 value:
	tex -> the texture ID of the texture that's just been created.

	Potential Errors:
	None...
	*/

		GLuint genTexture(SDL_Surface* surface) {
			GLenum textureFormat;
			GLint nOfColours;
			GLuint tex;

			//Watch out for people cheekily passing in NULL.
			if(surface) {
				//Make sure you're using the right colour scheme.
				nOfColours = surface->format->BytesPerPixel;
				if(nOfColours == 4){
                    #ifdef __APPLE__
                        textureFormat = GL_BGRA;
                    #else
                        textureFormat = GL_RGBA;
                    #endif
                }
				else if(nOfColours == 3)
					textureFormat = GL_RGB;
				else
					textureFormat = GL_NONE;

				//Generate and Bind the texture. The binding is done for the following manipulations.
				glGenTextures(1, &tex);
				glBindTexture(GL_TEXTURE_2D, tex);

				//Sets mip-map sampling type (i.e. bilinear, point).
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

				//Actualy create the texture.
				glTexImage2D(GL_TEXTURE_2D, 0, nOfColours, surface->w, surface->h, 0, textureFormat, GL_UNSIGNED_BYTE, surface->pixels);
				//Free the memory associated with the surface. Saves you from having to do it outside.
				SDL_FreeSurface(surface);
			}

			return tex;
		}
};

#define GFX_POOL GraphicPool::instance()

#endif // GRAPHICPOOL_HPP_INCLUDED
