#include <math.h>
#include "Gun.hpp"
#include "Position.hpp"
#include "Window.hpp"
#include "world.hpp"
#include "Sound.hpp"
#include "Input.hpp"
#include "GunList.hpp"

void Gun::init(GunListData data) {
	name_ = data.name;

	bangFile_ = data.bangFile;
	hitFile_ = data.hitFile;

	gunGfx_ = Sprite(data.gFileName, V64x64, Vector(0.0f, 0.0f), Vector(0.5f, 0.5f), 1);
	bulletGfx_ = Sprite(data.bFileName, V12x12, Vector(0.0f, 0.0f), Vector(1.0f, 1.0f), 1);

	recoil_ = data.recoil;
	shotLim_ = data.shotLim;
	life_ = data.life;

	fired_ = -1;
	effects_ = data.effects;
}

void Gun::update() {
	if(fired_ == 0) {
		gunGfx_.shunt(-1);
		--fired_;
	}
	else if(fired_ > 0)
		--fired_;

	gunGfx_.update();

	for(auto bullet = bullets_.begin(); bullet < bullets_.end();) {
		Vector oldVel(bullet->pos.vel());
		bullet->pos.update(1.0f);
		bullet->sprite.update();

		if(bullet->life > 0)
			--bullet->life;

		Block& current_block = CURRENT_ROOM.get_block(bullet->pos.pos() + oldVel);
		Prop* prop(CURRENT_ROOM.getProp(bullet->pos.pos() + oldVel));

		if(not current_block.is_traversable() || current_block.wantsEffect(effects_) || prop
       || bullet->pos.pos().x < 0 || bullet->pos.pos().x > 1.0
       || bullet->pos.pos().y < 0 || bullet->pos.pos().y > 1.0) {
			#ifndef __APPLE__
                AUDIO.open(hitFile_);
            #endif

			if(current_block.wantsEffect(effects_))
				CURRENT_ROOM.applyToBlock(CURRENT_ROOM.get_block_index(bullet->pos.pos() + oldVel), effects_ & current_block.wantedEffects());
			else if(prop) {
				if(prop->code() == 'y' || prop->code() == 'z')
					CURRENT_ROOM.applyToProp(CURRENT_ROOM.get_prop_index(bullet->pos.pos() + oldVel), EFF_BREAK, oldVel);
				else if(prop->wantsEffect(effects_))
					CURRENT_ROOM.applyToProp(CURRENT_ROOM.get_prop_index(bullet->pos.pos() + oldVel), effects_ & prop->wantedEffects(), oldVel);
			}

			bullet = bullets_.erase(bullet);
		}
		else if(bullet->life == 0)
			bullet = bullets_.erase(bullet);
		else
			++bullet;
	}
}

void Gun::draw(Position pos, bool drawGun) {
	for(std::vector<Bullet>::iterator bullet(bullets_.begin()); bullet < bullets_.end(); ++bullet)
		bullet->sprite.draw(bullet->pos.pos() - (bullet->sprite.dim() / 2.0f));

	if(drawGun) {
		Vector p[4];
		p[0] -= Vector(gunGfx_.dim().x * X_SHOULDER, gunGfx_.dim().y * Y_SHOULDER);
		p[1] -= Vector(gunGfx_.dim().x * X_SHOULDER, gunGfx_.dim().y * (Y_SHOULDER - 1.0f));
		p[2] -= Vector(gunGfx_.dim().x * (X_SHOULDER - 1.0f), gunGfx_.dim().y * (Y_SHOULDER - 1.0f));
		p[3] -= Vector(gunGfx_.dim().x * (X_SHOULDER - 1.0f), gunGfx_.dim().y * Y_SHOULDER);

		Vector diff(INPUT.x() - pos.pos().x - (gunGfx_.dim().x / 2.0f), INPUT.y() - pos.pos().y - (gunGfx_.dim().y / 2.0f));
		if(diff.x != 0.0f || diff.y != 0.0f) {
			float angle(acosf(diff.x / sqrtf((diff.x * diff.x) + (diff.y * diff.y))));
			//Hackery to fix the fact that cos only gives 0 -> 180
			if(diff.y < 0.0f)
				angle *= -1.0f;

			for(int i(0); i < 4; ++i)
				p[i] = Vector((p[i].x * cosf(angle)) - (p[i].y * sinf(angle)), (p[i].x * sinf(angle)) + (p[i].y * cosf(angle)));
		}

		for(int i(0); i < 4; ++i)
			p[i] += pos.pos() + Vector(gunGfx_.dim().x * X_SHOULDER, gunGfx_.dim().y * Y_SHOULDER);

		gunGfx_.draw(p[0], p[1], p[2], p[3]);
	}
}

bool Gun::shoot(Position aim) {
	if(bullets_.size() < shotLim_) {
		bullets_.push_back(Bullet(aim, bulletGfx_, life_));

		if(fired_ < 0)
			gunGfx_.shunt(1);
        #ifndef __APPLE__
            AUDIO.open(bangFile_);
        #endif

		fired_ = FIRING_TIME;
		return true;
	}
	else
		return false;
}
