#ifndef INPUT_HPP_INCLUDED
#define INPUT_HPP_INCLUDED

#include <SDL2/SDL.h>
#include <vector>
#include "Constants.hpp"
#include "Window.hpp"

/// A singleton to hold all of the inputs that happened since the last time it checked, so that anyone can reference them with ease.

class Input
{
	public:
	//Destructor
		~Input() {}

	//Instance fetcher. Static so it can be called without an object.
		static Input& instance() {
			static Input passInput;
			return passInput;
		}

	//Clears the queue of any inputs, and checks for any new ones.
		void scan() {
			events_.clear();
			SDL_Event event;

			while(SDL_PollEvent(&event))
				events_.push_back(event);

			int xMouse(0);
			int yMouse(0);
			SDL_GetMouseState(&xMouse, &yMouse);

			x_ = WINDOW.floatifyX(xMouse);
			y_ = WINDOW.floatifyY(yMouse);

			checkKeys();
		}

	//Returns the first event of the requested type that has happened.
		SDL_Event* event(unsigned int type) {
			for(std::vector<SDL_Event>::iterator itEvent(events_.begin()); itEvent < events_.end(); ++itEvent)
				if(itEvent->type == type)
					return &(*itEvent);

			return NULL;
		}

	//Checks for a specific input (for ease of access).
		bool quit() { return usefuls_[6] || event(SDL_QUIT); }

		bool up() { return usefuls_[0]; }
		bool down() { return usefuls_[1]; }
		bool left() { return usefuls_[2]; }
		bool right() { return usefuls_[3]; }

		bool space() { SDL_Event* hold(event(SDL_KEYDOWN)); return hold && hold->key.keysym.sym == SDLK_SPACE; }

		bool lMouse() { return usefuls_[4]; }
		bool rMouse() { return usefuls_[5]; }

		bool scrollUp() { SDL_Event* hold(event(SDL_MOUSEWHEEL)); return hold && hold->wheel.y > 0; }
		bool scrollDown() { SDL_Event* hold(event(SDL_MOUSEWHEEL)); return hold && hold->wheel.y < 0; }

	//Gets the x and y coordinates of the mouse.
		float x() { return x_; }
		float y() { return y_; }

	protected:
		//Constructor. Protected so no one else can make one.
		Input() {}

		//The vector storing the events that have happened.
		std::vector<SDL_Event> events_;

		//The mouse position (to save constant checking).
		float x_;
		float y_;

		bool usefuls_[7];

	private:
		void checkKeys() {
			for(std::vector<SDL_Event>::iterator itEvent(events_.begin()); itEvent < events_.end(); ++itEvent) {
				if(itEvent->type == SDL_KEYDOWN) {
					if(itEvent->key.keysym.sym == SDLK_UP || itEvent->key.keysym.sym == SDLK_w)
						usefuls_[0] = true;
					else if(itEvent->key.keysym.sym == SDLK_DOWN || itEvent->key.keysym.sym == SDLK_s)
						usefuls_[1] = true;
					else if(itEvent->key.keysym.sym == SDLK_LEFT || itEvent->key.keysym.sym == SDLK_a)
						usefuls_[2] = true;
					else if(itEvent->key.keysym.sym == SDLK_RIGHT || itEvent->key.keysym.sym == SDLK_d)
						usefuls_[3] = true;
					else if(itEvent->key.keysym.sym == SDLK_ESCAPE)
						usefuls_[6] = true;
				}
				else if(itEvent->type == SDL_KEYUP) {
					if(itEvent->key.keysym.sym == SDLK_UP || itEvent->key.keysym.sym == SDLK_w)
						usefuls_[0] = false;
					else if(itEvent->key.keysym.sym == SDLK_DOWN || itEvent->key.keysym.sym == SDLK_s)
						usefuls_[1] = false;
					else if(itEvent->key.keysym.sym == SDLK_LEFT || itEvent->key.keysym.sym == SDLK_a)
						usefuls_[2] = false;
					else if(itEvent->key.keysym.sym == SDLK_RIGHT || itEvent->key.keysym.sym == SDLK_d)
						usefuls_[3] = false;
					else if(itEvent->key.keysym.sym == SDLK_ESCAPE)
						usefuls_[6] = false;
				}
				else if(itEvent->type == SDL_MOUSEBUTTONDOWN) {
					if(itEvent->button.button == SDL_BUTTON_LEFT)
						usefuls_[4] = true;
					if(itEvent->button.button == SDL_BUTTON_RIGHT)
						usefuls_[5] = true;
				}
				else if(itEvent->type == SDL_MOUSEBUTTONUP) {
					if(itEvent->button.button == SDL_BUTTON_LEFT)
						usefuls_[4] = false;
					if(itEvent->button.button == SDL_BUTTON_RIGHT)
						usefuls_[5] = false;
				}
			}
		}
};

//Allows ease of fetching the instance.
#define INPUT Input::instance()

#endif // INPUT_HPP_INCLUDED
