#include <math.h>
#include <vector>
#include "Player.hpp"
#include "Constants.hpp"
#include "Position.hpp"
#include "Input.hpp"
#include "Sprite.hpp"
#include "Gun.hpp"
#include "world.hpp"
#include "Sound.hpp"
#include "GunList.hpp"

void Player::init() {
	sprite_.init("gfx/PC_Spritesheet.png", V64x64, Vector(), Vector(1.0f / 8.0f, 1.0f / 8.0f), 1);

	pos_.init(Vector(0.5f, 0.5f), Vector(), Vector(0.0f, GRAVITY));

	//guns_.clear();
	//for(int index(0); index < gunListLen; ++index)
	//	guns_.push_back(Gun(gunList[index]));

  gunMap_['c'] = 0;
  gunMap_['f'] = 1;
  gunMap_['d'] = 2;
  gunMap_['m'] = 3;
  gunMap_['g'] = 4;

	gunIndex_ = 0;//rand() % guns_.size();

	faceLeft_ = false;
	dead_ = false;
}

void Player::update() {
	//Moving and jumping.
	if(!WINDOW.fading()) {
		if(INPUT.left())
			pos_.addVel(Vector(-1.0f * PLAYER_SPEED, 0.0f));
		if(INPUT.right())
			pos_.addVel(Vector(PLAYER_SPEED, 0.0f));
		if(INPUT.up() && onFloor())
			pos_.setVel(Vector(pos_.vel().x, -1.0f * JUMP_SPEED));

		//Shooting at the mouse.
		if(INPUT.lMouse()) {
			Vector diff(Vector(INPUT.x(), INPUT.y()) - pos_.pos() - (sprite_.dim() / 2.0f));
			float len(sqrtf((diff.x * diff.x) + (diff.y * diff.y)));

			if(len != 0.0f && guns_[gunIndex_].shoot(Position(pos_.pos() + (sprite_.dim() / 2.0f), diff * BULLET_SPEED / len)))
				pos_.addVel(diff * -1.0f * guns_[gunIndex_].recoil() / len);
		}
	}

	//Changing weapon.
	if((INPUT.scrollUp() || INPUT.space()) && this->guns_.size() > 0)
		gunIndex_ = (gunIndex_ + 1) % guns_.size();
	else if(INPUT.scrollDown() && this->guns_.size() > 0)
		gunIndex_ = (gunIndex_ + guns_.size() - 1) % guns_.size();

	updatePos();
	updateSprite();

	for(std::vector<Gun>::iterator gun(guns_.begin()); gun < guns_.end(); ++gun)
		gun->update();
}

//Draw the guy's sprite.
void Player::draw() {
	if(faceLeft_)
		sprite_.draw(pos_);

	for(std::vector<Gun>::iterator gun(guns_.begin()); gun < guns_.end(); ++gun)
		gun->draw(Position(pos_, Vector(0.1f)), gun - guns_.begin() == gunIndex_);

	if(!faceLeft_)
		sprite_.draw(pos_);
}

void Player::kill()
{
	dead_ = true;
	WINDOW.fadeOut();

	pos_.setVel(Vector(0.0f, JUMP_SPEED / 2.0f) - pos_.vel());
}

void Player::updatePos() {
	pos_.update(VELOCITY_DECAY, Vector(PLAYER_TOP_LEFT_X * sprite_.dim().x, PLAYER_TOP_LEFT_Y * sprite_.dim().y),
  		Vector(PLAYER_BOTTOM_RIGHT_X * sprite_.dim().x, PLAYER_BOTTOM_RIGHT_Y * sprite_.dim().y));

	if((!CURRENT_ROOM.can_move_here(pos_.pos() + (sprite_.dim() / 2.0f)) || pos_.pos().y > 1.0f) && !WINDOW.fading())
		kill();

	if(!WINDOW.fading() && CURRENT_ROOM.get_block(pos_.pos() + (sprite_.dim() / 2.0f)).is_exit()
		and World::instance().num_rooms() == World::instance().room_limit()) {
		Game::instance().win();
		WINDOW.fadeOut();
	}
	else if(!WINDOW.fading() && CURRENT_ROOM.get_block(pos_.pos() + (sprite_.dim() / 2.0f)).is_exit())
		WINDOW.fadeOut();
	if(WINDOW.faded()) {
		pos_.setPos(CURRENT_ROOM.get_entry_block()->pos());
		pos_.setVel(Vector());
	}
}

void Player::updateSprite() {
	if(INPUT.x() < pos_.pos().x)
		faceLeft_ = true;
	else if(INPUT.x() > pos_.pos().x)
		faceLeft_ = false;

	if((faceLeft_ != (pos_.vel().x < 0) && sprite_.step() != -1) || (faceLeft_ == (pos_.vel().x < 0) && sprite_.step() == -1))
		sprite_.reverse();

	if(dead_) {
		if(faceLeft_)
			setGFXStuff(7.0f / 8.0f, 1, 1);
		if(!faceLeft_)
			setGFXStuff(3.0f / 8.0f, 1, 0);
	}
	else if(onFloor()) {
		sprite_.texPos().x = 0.0f;

		if(pos_.vel().x == 0.0f) {
			if(faceLeft_)
				setGFXStuff(0.5f, 1, 1);
			if(!faceLeft_)
				setGFXStuff(0.0f, 1, 0);
		}
		else if(faceLeft_)
			setGFXStuff(0.625f, 8, 1);
		else if(!faceLeft_)
			setGFXStuff(0.125f, 8, 0);
	}
	else {
		if(faceLeft_)
			setGFXStuff(0.75f, 1, 1);
		else
			setGFXStuff(0.25f, 1, 0);

		if(pos_.vel().y > 0.0f)
			sprite_.texPos().x = sprite_.texDim().x;
		else if(pos_.vel().y < 0.0f)
			sprite_.texPos().x = 0.0f;
	}

	sprite_.update();
}

void Player::setGFXStuff(float texPosY, int frames, int row) {
	sprite_.texPos().y = texPosY;
	sprite_.setFrames(frames);

  if (guns_.size() > 0)
	  guns_[gunIndex_].setRow(row);
}

bool Player::onFloor() {
	return not CURRENT_ROOM.can_move_here(pos_.pos() + Vector(sprite_.dim().x / 2.0f, CURRENT_ROOM.get_block_height()));
}
