#include <cstdlib>
#include <SDL.h>
#include <SDL_ttf.h>
#include "Constants.hpp"
#include "GraphicPool.hpp"
#include "Window.hpp"
#include "Position.hpp"

Window::~Window()
{
	SDL_GL_DeleteContext(context_);
	close();
}

void Window::open(std::string name, int w, int h, Uint32 flags)
{
	name_ = name;
	w_ = w;
	h_ = h;

	//Initialising SDL.
	SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO);
	atexit(SDL_Quit);

	window_ = SDL_CreateWindow(name.c_str(), SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, w, h, flags | SDL_WINDOW_OPENGL);

	//Creating the context for openGL to draw in if necessary.
	if(!contextExists_) {
		context_ = SDL_GL_CreateContext(window_);
		contextExists_ = true;
	}

	//Initialising openGL.
	glClearColor(0, 0, 0, 0);
	glClearDepth(1.0f);

	glViewport(0, 0, w, h);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	//Draw in a box with no depth scaling, from (0, 0) = top left corner, to (1, 1) = bottom right corner.
	glOrtho(0, 1, 1, 0, 1, -1);
	glMatrixMode(GL_MODELVIEW);

	//To allow images to use the alpha channel.
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glLoadIdentity();

	createBlank();
	createFont();
	bgTex_ = GFX_POOL.loadImage("gfx/Background_Tile.png");
}

void Window::close() { SDL_DestroyWindow(window_); }

int Window::getWidth() { return w_; }
int Window::getHeight() { return h_; }

void Window::drawTri(Vector p1, Vector p2, Vector p3, Colour colour)
{
	glBegin(GL_TRIANGLES);
		glColor4f(colour.r, colour.g, colour.b, colour.a);
		glVertex2f(p1.x, p1.y);
		glVertex2f(p2.x, p2.y);
		glVertex2f(p3.x, p3.y);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glEnd();
}

void Window::draw(GLuint tex, Vector p1, Vector p2, Vector p3, Vector p4, Vector tPos, Vector tDim, Colour palette)
{
	//Getting ready to use textures.
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, tex);

	//Drawing the quad.
	glBegin(GL_QUADS);
		glColor4f(palette.r, palette.g, palette.b, palette.a);
		glTexCoord2f(tPos.x, tPos.y);
		glVertex2f(p1.x, p1.y);
		glTexCoord2f(tPos.x, tPos.y + tDim.y);
		glVertex2f(p2.x, p2.y);
		glTexCoord2f(tPos.x + tDim.x, tPos.y + tDim.y);
		glVertex2f(p3.x, p3.y);
		glTexCoord2f(tPos.x + tDim.x, tPos.y);
		glVertex2f(p4.x, p4.y);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glEnd();

	//Then, stopping using textures. This is done so that there's no accidental overlap with pallettes if I draw something non-texture based later.
	glDisable(GL_TEXTURE_2D);
}

//A function to draw some text on the screen, in a given colour, at the given point.
void Window::print(std::string text, Vector pos, Colour palette, float fontSize)
{
	for(int index(0); index < int(text.size()); ++index)
		draw(font_, Vector(pos.x + (index * fontSize), pos.y), Vector(fontSize, fontSize * floatifyY(w_)), Vector((text[index] - ' ') * ASCII_SPACING, 0.0f), Vector(ASCII_SPACING, 1.0f), palette);
}

//The function to update the screen with the contents of the buffer.
void Window::flush()
{
	//Draw the fading if it's necessary.
	fade_ += fadeStep_;
	if(fade_ <= 0.0f)
		fadeStep_ = 0.0f;
	else {
		if(fade_ >= 1.0f)
			fadeStep_ *= -1.0f;

		drawTri(Vector(), Vector(0.0f, 1.0f), Vector(1.0f, 0.0f), Colour(0.0f, 0.0f, 0.0f, fade_));
		drawTri(Vector(1.0f, 1.0f), Vector(0.0f, 1.0f), Vector(1.0f, 0.0f), Colour(0.0f, 0.0f, 0.0f, fade_));
	}

	//Make the stuff in the buffer visible.
	glFlush();
	SDL_GL_SwapWindow(window_);

	//Then clear the buffer for the next load of stuff.
	glClear(GL_COLOR_BUFFER_BIT);
	glLoadIdentity();

	draw(bgTex_, Vector(), Vector(1.0f, 1.0f), Vector(0.0f, 0.0f), Vector(1.0f, 1.0f));
}

GLuint Window::getBlank() { return blankTex_; }

//Make the screen fade out, and check if it's done so.
void Window::fadeOut() { fadeStep_ = FADE_SPEED; }
bool Window::fading() { return fade_ > 0.0f; }
bool Window::faded() { return fade_ >= 1.0f; }

//A function to convert an absolute int into a relative float.
float Window::floatifyX(int in) { return float(in) / float(w_); }
float Window::floatifyY(int in) { return float(in) / float(h_); }

//Create a clear texture for use as the default texture.
void Window::createBlank()
{
	SDL_Surface* blankSurface(SDL_CreateRGBSurface(0, 1, 1, 32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0xff000000));
	blankTex_ = GFX_POOL.loadImage("0", blankSurface);
}

//Create the font and texturify it.
void Window::createFont()
{
	//Create the font.
	TTF_Init();
	TTF_Font* font(TTF_OpenFont("gfx/FreeMono.ttf", int(1.5f * TEXT_SIZE * h_)));

	//Set the colour (r, g, b).
	SDL_Colour colour = { 255, 255, 255 };

	//Print the string of ASCII characters, then create an empty surface, then blit the ASCII onto the empty surface, then delete the ASCII surface (passSurface is deleted by genTexture).
	SDL_Surface* textSurface(TTF_RenderText_Solid(font, " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~", colour));
	SDL_Surface* passSurface(SDL_CreateRGBSurface(0, textSurface->w, textSurface->h, 32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0xff000000));
	SDL_BlitSurface(textSurface, 0, passSurface, 0);
	SDL_FreeSurface(textSurface);

	//Ask GraphicPool to generate the texture.
	font_ = GFX_POOL.loadImage("Text", passSurface);

	//Clean up after meself.
	TTF_CloseFont(font);
}
