#ifndef WINDOW_HPP_INCLUDED
#define WINDOW_HPP_INCLUDED

#include <SDL.h>
#ifdef __APPLE__
	#include <OpenGL/gl.h>
#else
	#include <GL/gl.h>
#endif
#include "Position.hpp"

struct Colour {
	float r, g, b, a;
	Colour (float rIn = 1.0f, float gIn = 1.0f, float bIn = 1.0f, float aIn = 1.0f) :r(rIn), g(gIn), b(bIn), a(aIn) {}
};
#define WHITE Colour()
#define BLACK Colour(0.0f,0.0f,0.0f)

class Window
{
	public:
	//Destructor. Calls close() in case I want anything to happen for certain when it closes.
		~Window();

	//The instance retriever. Uses static to ensure that only one exists during execution.
		static Window& instance() {
			static Window passInstance;
			return passInstance;
		}

/*
The function to open a new window. This should not be called if one is already open (for safety).

Takes 3 parameters:
	name -> the name of the window. This gets put on the bar across the top.
	w, h -> the dimensions of the window in pixels.

Potential Errors:
	w and h are <= 0.
*/

		void open(std::string name, int w, int h, Uint32 flags = 0);

	//The function that gets called when the window is closed.
		void close();

	//Functions to get the dimensions of the window in pixels.
		int getWidth();
		int getHeight();

	//A function to draw a monocoloured triangle.
		void drawTri(Vector p1, Vector p2, Vector p3, Colour colour);

/*
The function to draw the given texture on a rectangle of dimension as given, at the position given.

Takes 3-6 parameters:
	tex -> the ID of the textre to be loaded and drawn.
	pos -> the position (relative (0-1)) of the top left corner of the rectangle.
	dim -> the dimensions (relative (0-1)) of the rectangle.

	tPos, tDim -> the coordinates of the bit of the texture to draw.
	palette -> a Colour object (0-1, 0-1, 0-1, 0-1) that stores the colour to tinge the texture.

Potential Errors:
	tex does not associate with a texture.
*/
		void draw(GLuint tex, Vector p1, Vector p2, Vector p3, Vector p4, Vector tPos, Vector tDim, Colour palette = WHITE);

	//The version for drawing a square.
		void draw(GLuint tex, Vector pos, Vector dim, Vector tPos = Vector(), Vector tDim = Vector(1.0f, 1.0f), Colour palette = WHITE)
			{ draw(tex, pos, pos + Vector(0.0f, dim.y), pos + dim, pos + Vector(dim.x, 0.0f), tPos, tDim, palette); }

	//The version with just a palette.
		void draw(GLuint tex, Vector pos, Vector dim, Colour palette) { draw(tex, pos, dim, Vector(), Vector(1.0f, 1.0f), palette); }

	//A function to draw some text on the screen, in a given colour, at the given point.
		void print(std::string text, Vector pos, Colour palette = BLACK, float fontSize = TEXT_SIZE);

	//The function to update the screen with the contents of the buffer.
		void flush();

		GLuint getBlank();

	//Make the screen fade out, and check if it's done so.
		void fadeOut();
		bool fading();
		bool faded();

	//A function to convert an absolute int into a relative float.
		float floatifyX(int in);
		float floatifyY(int in);

	protected:
		Window () { contextExists_ = false;}

		SDL_Window* window_;
		SDL_GLContext context_;
		bool contextExists_;

		std::string name_;
		int w_;
		int h_;

		GLuint blankTex_;
		GLuint font_;
		GLuint bgTex_;

		float fade_;
		float fadeStep_;

	private:
		//Create a clear texture for use as the default texture.
		void createBlank();

		//Create the font and texturify it.
		void createFont();
};

#define WINDOW Window::instance()
#define TEX_0 WINDOW.getBlank()

#endif // WINDOW_HPP_INCLUDED
