#include <fstream>
#include <sstream>
#include <random>

#include "room.hpp"
#include "Sound.hpp"
#include "PropList.hpp"

Room::Room(Room&& room)
  : block_width_(std::move(room.block_width_)),
    block_height_(std::move(room.block_height_)),
    entry_block_(std::move(room.entry_block_)),
    exit_block_(std::move(room.exit_block_)),
    random_blocks_(std::move(room.random_blocks_)),
    props_(std::move(room.props_)),
    propIter_(std::move(room.propIter_)),
    gate_(std::move(room.gate_)),
    gateInRoom_(std::move(room.gateInRoom_)),
    empty_(std::move(room.empty_))
{
  // Move constructor
  this->blocks_.resize(room.blocks_.size());
  for (int i = 0; i < this->blocks_.size(); ++i) {
    this->blocks_[i] = BlockPtr(std::move(room.blocks_[i]));
  }
}


Room& Room::operator=(Room&& room)
{
  this->block_height_ = std::move(room.block_height_);
  this->block_width_ = std::move(room.block_width_);
  this->entry_block_ = std::move(room.entry_block_);
  this->exit_block_ = std::move(room.exit_block_);
  this->random_blocks_ = std::move(room.random_blocks_);
  this->props_ = std::move(room.props_);
  this->propIter_ = std::move(room.propIter_);
  this->gate_ = std::move(room.gate_);
  this->gateInRoom_ = std::move(room.gateInRoom_);
  this->empty_ = std::move(room.empty_);

  this->blocks_.resize(room.blocks_.size());
  for (int i = 0; i < this->blocks_.size(); ++i) {
    this->blocks_[i] = BlockPtr(std::move(room.blocks_[i]));
  }
  return *this;
}


template <typename T>
void Room::add_block(const T& block)
{
  this->blocks_.push_back(std::unique_ptr<Block>(new T(block)));
}


template <typename T>
void Room::replace_block(const int index) {
  // Replace the specified tile with the given one
  this->blocks_[index] = BlockPtr(new T(this->blocks_[index]->pos()));
}
template void Room::replace_block<AirBlock>(const int);

void Room::applyToBlock(int index, int effect, const Vector& vec)
{
	if(effect & EFF_BREAK || effect & EFF_DIG)
		replace_block<AirBlock>(index);
	else if(effect & EFF_FREEZE)
		replace_block<IceBlock>(index);
}

void Room::applyToProp(int index, int effect, const Vector& vec)
{
	if(effect & EFF_BREAK || effect & EFF_DIG || effect & EFF_FREEZE)
		props_[index].change(RES_DOOMED);
	else if(effect & EFF_DRAIN)
		props_[index].change(RES_ADD_GRAV);
	else if (effect & EFF_MOVE)
		props_[index].pos().setVel(Vector(0.01 * ((vec.x > 0) ? 1.0 : -1.0), 0.0));
}

void Room::make_from_file(const std::string &path)
{
  std::cout << "Info: Loading room file " << path << std::endl;
  gateInRoom_ = false;

  std::vector<std::string> file_lines;
  std::ifstream input_file(path, std::ifstream::in);

  std::string line;
  // Grab file contents
  while (std::getline(input_file, line)) {
    file_lines.push_back(line);
  }

  // Compute block height

  unsigned long world_width = file_lines[0].size();
  unsigned long world_height = file_lines.size();
  this->block_height_ = 1.0f / world_height;
  this->block_width_ = 1.0f / world_width;

  Vector current_coords;
  this->blocks_.clear();
  int index = 0;
  for (auto& line : file_lines) {
    current_coords.x = 0.0;
    // Error checking
    if (line.size() != world_width) {
      //std::cout << "Error! Lines in room file aren't same length." << std::endl;
    }
    // Loop through line and add blocks
    for (auto ch : line) {
      BlockPtr new_block_ptr;
      // Could we use a std::map here?
      switch (ch) {
        case ' ':
          new_block_ptr = BlockPtr(new AirBlock(Position(current_coords)));
          break;
        case '*':
          new_block_ptr = BlockPtr(new BoundaryBlock(Position(current_coords)));
          break;
        case 'E':
        case 'S':
		new_block_ptr = BlockPtr(new EntranceBlock(Position(current_coords)));
		this->entry_block_ = new_block_ptr.get();

		gate_.init(propList[2], Position(current_coords));
		gateInRoom_ = true;

		break;
        case 'X':
        case 'F':
          new_block_ptr = BlockPtr(new ExitBlock(Position(current_coords)));
          this->exit_block_ = new_block_ptr.get();
          break;
        case 'B':
          new_block_ptr = BlockPtr(new Block(Position(current_coords)));
          break;
        case 'D':
          new_block_ptr = BlockPtr(new DirtBlock(Position(current_coords)));
          break;
        case 'W':
          new_block_ptr = BlockPtr(new WaterBlock(Position(current_coords)));
          break;
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
          this->random_blocks_.push_back(std::vector<int>{index, atoi(&ch)});
          new_block_ptr = BlockPtr(new AirBlock(Position(current_coords)));
          break;
        default:
		      bool found(false);
		      for(int index(0); index < propListLen && !found; ++index) {
		  	    if(propList[index].code == ch) {
			  	    props_.push_back(Prop(propList[index], Position(current_coords, Vector(), Vector(0.0f, GRAVITY))));
				      found = true;
			      }
		      }

		      new_block_ptr = BlockPtr(new Block(Position(current_coords)));
          break;
      }
      this->blocks_.push_back(std::move(new_block_ptr));
      current_coords.x += this->block_width_;
      index++;
    }
    current_coords.y += this->block_height_;
  }
}


Block& Room::get_block(const Position &coordinates) {
  // Loop through blocks and find out which one is at specified coordinates

  for (auto& block : this->blocks_) {
    if (coordinates.pos().x >= block->pos().x
      and coordinates.pos().x < block->pos().x + this->block_width_
      and coordinates.pos().y >= block->pos().y
      and coordinates.pos().y < block->pos().y + this->block_height_) {
      return *block;
    }
  }
  //std::cout << "Warning! Block at (" << coordinates.pos().x << ", " << coordinates.pos().y << ") not found." << std::endl;
  //Game::instance().getGuy()->kill();
  return this->empty_;
}


const Block& Room::get_block(const Position &coordinates) const {
  // Loop through blocks and find out which one is at specified coordinates

  for (auto& block : this->blocks_) {
    if (coordinates.pos().x >= block->pos().x
      and coordinates.pos().x < block->pos().x + this->block_width_
      and coordinates.pos().y >= block->pos().y
      and coordinates.pos().y < block->pos().y + this->block_height_) {
      return *block;
    }
  }
  //std::cout << "Warning! Block at (" << coordinates.pos().x << ", " << coordinates.pos().y << ") not found." << std::endl;
  //Game::instance().getGuy()->kill();
  return this->empty_;
}

Prop* Room::getProp(Position coords)
{
	if(gateInRoom_ && gate_.inside(coords.pos()))
		return &gate_;

	for(std::vector<Prop>::iterator prop(props_.begin()); prop < props_.end(); ++prop)
		if(prop != propIter_ && prop->inside(coords.pos()))
			return &(*prop);
  //Game::instance().getGuy()->kill();
	return NULL;
}

int Room::get_block_index(const Position &coordinates) {
  // Loop through blocks and find out which one is at specified coordinates

  for (int index(0); index < this->blocks_.size(); ++index) {
    if (coordinates.pos().x >= blocks_[index]->pos().x
      and coordinates.pos().x < blocks_[index]->pos().x + this->block_width_
      and coordinates.pos().y >= blocks_[index]->pos().y
      and coordinates.pos().y < blocks_[index]->pos().y + this->block_height_) {
      return index;
    }
  }

  //std::cout << "Warning! Block at (" << coordinates.pos().x << ", " << coordinates.pos().y << ") not found." << std::endl;
  //Game::instance().getGuy()->kill();
  return 0;
}

int Room::get_prop_index(Position coords)
{
	for(int index(0); index < int(props_.size()); ++index)
		if(props_[index].inside(coords.pos()))
			return index;

	return -1;
}

bool Room::can_move_here(const Position &coordinates)
{
	Prop* hold(getProp(coordinates));
	if(hold && hold->solid())
		return false;

  // Determine if the current block is traversable
  Block& block = this->get_block(coordinates);
  return block.is_traversable();
}

int Room::tiling_index(int i) {
    
    int x = i % ROOM_SIZE_X;
    int y = (i - x)/ROOM_SIZE_X;
    
    int id = this->blocks_[ y * ROOM_SIZE_X + x ]->sprite_index();
    
    //Check up.
    int tilingIndex(0);
    if (check_adjacent(id, x, y - 1))
        tilingIndex |= 1;
    
    //Check right.
    if (check_adjacent(id, x + 1, y))
        tilingIndex |= 2;
    
    //Check down.
    if (check_adjacent(id, x, y + 1))
        tilingIndex |= 4;
    
    //Check left.
    if (check_adjacent(id, x - 1, y))
        tilingIndex |= 8;
    
    return tilingIndex;
}

bool Room::check_adjacent(int checkID, int x, int y) {
    
    //If coordinates outside room bounds, return true.
    if (x < 0 || x >= ROOM_SIZE_X || y < 0 || y >= ROOM_SIZE_Y)
        return false;
    
    //If ID of target cell equal to check ID, return true.
    if (this->blocks_[ y * ROOM_SIZE_X + x ]->sprite_index() == checkID)
        return false;
    
    //Otherwise, return false.
    return true;
}

void Room::update()
{
  std::random_device rd;
  std::mt19937 generator(rd());
  std::uniform_int_distribution<> run_dist(0, 20);
  if (this->random_blocks_.size() > 0 and run_dist(generator) == 0) {
    std::uniform_int_distribution<> random_block_dist(0, random_blocks_.size() - 1);
    std::uniform_int_distribution<> pass_fail_dist(0, 10);
    int block_int = random_block_dist(generator);
    auto &random_block = this->random_blocks_[block_int];

    if (pass_fail_dist(generator) < random_block[1]) {
      Vector &coords = this->blocks_[random_block[0]]->pos();
      props_.push_back(Prop(propList[0], coords));
      this->random_blocks_.erase(this->random_blocks_.begin() + block_int);
    }
  }

	if(gateInRoom_) {
		gate_.update();
		if(!gate_.solid() && !gate_.inside(Game::instance().getGuy()->pos().pos()) && !gate_.inside(Game::instance().getGuy()->pos().pos() + Game::instance().getGuy()->sprite()->dim())) {
			#ifndef __APPLE__
				AUDIO.open("audio/swish.wav");
			#endif
			gate_.sprite().texPos().x = gate_.sprite().texDim().x;
			gate_.solidify();
		}
	}

	for(propIter_ = props_.begin(); propIter_ < props_.end();) {
		propIter_->update();

		if(propIter_->changed() && propIter_->result() == RES_DOOMED)
			propIter_ = props_.erase(propIter_);
		else
			++propIter_;
	}
}

void Room::draw()
{
	Vector x_offset(block_width_, 0.0);
	Vector y_offset(0.0, block_height_);

	for(int index(0); index < blocks_.size(); ++index) {
		if(blocks_[index]->sprite_index() > -1) {
			blocks_[index]->sprite().texPos().x = tiling_index(index) * blocks_[index]->sprite().texDim().x;
			blocks_[index]->sprite().draw(blocks_[index]->pos());
		}
	}

	if(gateInRoom_)
		gate_.draw();

	for(std::vector<Prop>::iterator prop(props_.begin()); prop < props_.end(); ++prop)
		prop->draw();
}
